
from abstract_webtools import *
from .imports import *
# one call, one object, passed everywhere — no module-level globals
config = AppConfig.from_env()
MAX_FITZ_PAGES = 2000   # safety ceiling — don't OCR enormous PDFs without explicit override
def path_to_url(path,root_url=None,media_root=None):
    media_root = media_root or MEDIA_ROOT
    root_url = root_url or ROOT_URL
    return path.replace(media_root,root_url)
def url_to_path(path,root_url=None,media_root=None):
    media_root = media_root or MEDIA_ROOT
    root_url = root_url or ROOT_URL
    return path.replace(root_url,media_root)
def get_parsed_url(domain, **kwargs):
    parsed_url = dict(kwargs)
    post_variants = []
    # http / www
    http_www = get_http_www(domain)
    parsed_url.update(http_www)
    http = http_www.get('http')
    # basic domain pieces
    domain_paths = get_domain_paths(domain, http=http)
    if 'path' not in parsed_url:
        parsed_url['path']=[]
    parsed_url['path']+=domain_paths
    domain_name_ext = get_domain_name_ext(domain, http=http)
    parsed_url.update(domain_name_ext)

    domain_name = parsed_url.get('name',"") or ""
    domain = parsed_url.get('domain',"") or ""
        # tokenization
    tokenized_domain = tokenize_domain(domain)
    parsed_url["tokenized_domain"] = tokenized_domain
    app_name = " ".join(tokenized_domain)
    
    parsed_url["app_name"] = app_name
    # author / "i_url"
    parsed_url["author"] = f"@{domain_name.lower()}"
    parsed_url["i_url"] = f"{domain_name}://"

    # combine with domain
    # compute final title
    title = get_title(parsed_url)
    
    post_variants=[title,app_name,domain]
    variants = title_variants_from_domain(domain)
    base_variants = list(set([variant for variant in variants if variant not in post_variants]))
    # update the organized variants
    parsed_url["title_variants"] = get_all_title_variants(variants=base_variants,page=title,name=app_name,domain=domain)

    parsed_url["title"] = pad_or_trim(
        "title",
        string=title,
        title_variants=parsed_url["title_variants"],
        page=title,
        domain=domain,
        name = app_name
    )
    # get keywords
    keywords_info = get_keywords(parsed_url,page=title,domain=domain,name = app_name)
    parsed_url.update(keywords_info)
    keywords = parsed_url.get("keywords", [])
    # FINAL: longest→shortest list with TITLE first, DOMAIN second
    domain = parsed_url.get("domain")
    if domain:
        final_variants = [title,page_data ]
        # remove title/domain from pool
        pool = set(keywords + variants)
        pool.discard(title)
        pool.discard(parsed_url.get("domain"))

        # sort longest → shortest
        final_variants += sort_longest_first(pool)

        parsed_url["title_variants"] = final_variants
   
    return parsed_url
def init_page_data(config):
    page_data = {}
    ROOT_URL = config.root_url
    MEDIA_ROOT = config.media_root
    SITE_NAME = config.site_name
    DOMAIN = config.domain
    variants=title_variants_from_domain(DOMAIN)
    title_variants=title_variants_from_domain(DOMAIN) 
    site_name=SITE_NAME
    domain=DOMAIN
    root_url=ROOT_URL
    page_data = getInfo(domain=DOMAIN)
    page_data = get_parsed_url(**page_data)
    page_data['domain']=domain
    page_data['variants']=variants
    page_data['site'] = root_url
    page_data['site_name'] = site_name
    page_data['creator'] = site_name
    page_data['author'] = f'@{site_name}'
    return page_data

def get_page_data(
    title,
    href,
    summary,
    keywords,
    keywords_str,
    thumbnail,
    alt=None,
    caption=None,
    ):
    config = SiteConfig.from_env()
    ROOT_URL = config.root_url
    MEDIA_ROOT = config.media_root
    thumbnail_url=path_to_url(thumbnail,ROOT_URL,MEDIA_ROOT)
    page_data = init_page_data(config)
    page_data["title"] = title
    page_data['href'] = href
    page_data["page_url"] = href
    page_data['share_url'] = href
    page_data['thumbnail'] = thumbnail
    page_data['thumbnail_link'] = thumbnail_url
    page_data['description'] = summary
    page_data['keywords'] = keywords
    page_data['keywords_str']= ','.join(keywords)
    page_data["alt"] = alt or title
    page_data["caption"] = caption or summary
    meta_info = get_meta_info(info=page_data,base_image_dir=MEDIA_ROOT,base_image_url=ROOT_URL)
    meta_info['thumbnail_resized'] = path_to_url(meta_info.get('thumbnail_resized',thumbnail_url) or thumbnail_url,media_root=MEDIA_ROOT,root_url=ROOT_URL)
    meta_info['thumbnail_url_resized'] = path_to_url(meta_info.get('thumbnail_url_resized',thumbnail_url) or thumbnail_url,media_root=MEDIA_ROOT,root_url=ROOT_URL)
    meta_info['thumbnail'] = path_to_url(meta_info.get('thumbnail',thumbnail_url) or thumbnail_url,media_root=MEDIA_ROOT,root_url=ROOT_URL)
    meta_info['og']['image'] = path_to_url(meta_info.get('og',{}).get('image',thumbnail_url) or thumbnail_url,media_root=MEDIA_ROOT,root_url=ROOT_URL)
    meta_info['twitter']['image'] = path_to_url(meta_info.get('twitter',{}).get('image',thumbnail_url) or thumbnail_url,media_root=MEDIA_ROOT,root_url=ROOT_URL)

    return meta_info
##    page_data["alt"] = alt or title
##    page_data["caption"] = caption or summary
##    page_data["schema"] = meta.get("og", {})
##    page_data["schema"]['site_name'] = SITE_NAME
##    page_data["social_meta"] = meta.get("twitter", {})
##    page_data["other"] = meta.get("other")
##    page_data['meta']= meta
##    return page_data



from ..imports import *
def get_audio_duration(file_path):
    audio = get_pydub("AudioSegment").from_wav(file_path)
    duration_seconds = len(audio) / 1000
    duration_formatted = format_timestamp(len(audio))
    return duration_seconds,duration_formatted

def pick_optimal_thumbnail(whisper_result, keywords,thumbnail_paths=None,directory=None, *args,initial=None,end=None,**kwargs):
    scores = []
    keywords = keywords or []
    dirbase = os.path.basename(os.path.dirname(directory))
    image_files = thumbnail_paths
    if not thumbnail_paths:
        
        image_files = sort_frames(directory=directory)
##    image_files = clip_frames(image_files)
    first_image_file = image_files[0]
    filename,thumb_ext = os.path.splitext(first_image_file)
    
##    thumb_name,thumb_ext = os.path.splitext(os.listdir(directory)[0])
    # Process each Whisper segment
    for segment in whisper_result["segments"]:
        text = segment["text"].lower().strip()
        start_time = segment["start"]        # Find the closest thumbnail based on start time
        frame_number = math.floor(start_time)
        thumbnail_name = f"{dirbase}_frame_{frame_number}{thumb_ext}"
        
        # Check if thumbnail exists
        if thumbnail_name not in image_files:
            continue
        
        # Score the caption
        keyword_score = sum(1 for kw in keywords if kw.lower() in text)
        clarity_score = 1 if len(text) > 20 else 0
        end_phrase_penalty = -1 if "thanks for watching" in text else 0
        total_score = keyword_score + clarity_score + end_phrase_penalty
        
        # Store thumbnail path, score, and caption
        thumbnail_path = os.path.join(directory, thumbnail_name)
        scores.append((thumbnail_path, total_score, text))
       
    # Sort by score (highest first)
    scores = sorted(scores, key=lambda x: x[1], reverse=True)
    return scores[0] if scores else None

def extract_video_frames(video_path,directory,video_id=None,frame_interval=None):
    frame_interval = frame_interval or 1
    video = get_moviepy("VideoFileClip")(video_path)
    duration = video.duration
    video_id = video_id or generate_file_id(video_path)
    for t in range(0, int(duration), frame_interval):
        frame_path = os.path.join(directory,f"{video_id}_frame_{t}.jpg")
        if not os.path.isfile(frame_path):
            frame = video.get_frame(t)
            cv2.imwrite(frame_path, cv2.cvtColor(np.array(frame), cv2.COLOR_RGB2BGR))

def get_video_metadata(file_path):
    video = get_moviepy("VideoFileClip")(file_path)
    
    metadata = {
        'resolution': f"{video.w}x{video.h}",
        'format': 'MP4',
        'file_size_mb': os.path.getsize(file_path) / (1024 * 1024)
    }
    
    video.close()
    return metadata


def preprocess_for_ocr(image_path: str) -> np.ndarray:
    img = cv2.imread(image_path)
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    # Enhance contrast
    clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
    contrast = clahe.apply(gray)

    # Denoise
    denoised = cv2.bilateralFilter(contrast, d=9, sigmaColor=75, sigmaSpace=75)

    # Thresholding
    thresh = cv2.adaptiveThreshold(
        denoised,
        255,
        cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
        cv2.THRESH_BINARY_INV,
        blockSize=11,
        C=2
    )
    # Sharpen
    kernel = np.ones((3, 3), np.uint8)
    morph = cv2.dilate(thresh, kernel, iterations=1)    
    sharpen_kernel = np.array([[-1, -1, -1], [-1, 9, -1], [-1, -1, -1]])
    sharpened = cv2.filter2D(thresh, -1, sharpen_kernel)

    return sharpened

def extract_text_from_image(image_path: str) -> str:
    try:
        processed_img = preprocess_for_ocr(image_path)
        pil_img = Image.fromarray(cv2.bitwise_not(processed_img))  # invert for OCR
        text = pytesseract.image_to_string(pil_img, lang='eng')
        return text
    except Exception as e:
        print(f"[OCR Error] {e}")
        return ""

def is_frame_analyzed(frame_file,video_text_data):
    for values in video_text_data:
        frame = values
        if isinstance(values,dict):
            frame = values.get("frame")
        if frame_file == frame:
            return True

def extract_image_text(image_path,remove_phrases=None):
    remove_phrases = remove_phrases or []
    if is_media_type(image_path,media_types=['image']):
        raw_text = extract_text_from_image(image_path)
        cleaned_text = clean_text(raw_text)
        text = determine_remove_text(cleaned_text,remove_phrases=remove_phrases)
        return text

def extract_text_from_frame(image_path,image_texts,remove_phrases=None):
    basename = os.path.basename(image_path)
    if not is_frame_analyzed(basename,image_texts):
        text = extract_image_text(image_path,remove_phrases)
        if text:
            image_texts.append( {"frame": basename, "text": text})
    return image_texts

def extract_image_texts_from_directory(directory,image_texts=None,remove_phrases=None):
    image_texts = image_texts or []
    image_files = get_all_file_types(types=['image'],directory=directory)
    for i,image_path in enumerate(image_files):
        image_texts = extract_text_from_frame(image_path,image_texts)
        
    image_texts = sort_frames(image_texts)
    return image_texts


def determine_remove_text(text,remove_phrases=None):
    remove_phrases=remove_phrases or []
    found = False
    for remove_phrase in remove_phrases:
        if remove_phrase in text:
            found = True
            break
    if found == False:
        return text




def getPercent(i):
    return divide_it(i, 100)


def determine_remove_text(text,remove_phrases=None):
    remove_phrases=remove_phrases or []
    found = False
    for remove_phrase in remove_phrases:
        if remove_phrase in text:
            found = True
            break
    if found == False:
        return text


def generate_file_id(path: str, max_length: int = 50) -> str:
    base = os.path.splitext(os.path.basename(path))[0]
    base = unicodedata.normalize('NFKD', base).encode('ascii', 'ignore').decode('ascii')
    base = base.lower()
    base = re.sub(r'[^a-z0-9]+', '-', base).strip('-')
    base = re.sub(r'-{2,}', '-', base)
    if len(base) > max_length:
        h = hashlib.sha1(base.encode()).hexdigest()[:8]
        base = base[: max_length - len(h) - 1].rstrip('-') + '-' + h
    return base

def clean_text(text: str) -> str:
    text = re.sub(r'\s+', ' ', text)
    text = re.sub(r'[^a-zA-Z0-9\s:.,-]', '', text)
    text = text.strip()
    return text

def get_frame_number(file_path):
    if isinstance(file_path,dict):
        file_path = file_path.get('frame')
        
    file_path = '.'.join(file_path.split('.')[:-1])
    return int(file_path.split('_')[-1])

def sort_frames(frames=None,directory=None):
    if frames in [None,[]] and directory and os.path.isdir(directory):
        frames = get_all_file_types(types=['image'],directory=directory)
    frames = frames or []
    
    frames = sorted(
        frames,
        key=lambda x: get_frame_number(x) 
    )
    return frames
    

def get_from_list(list_obj=None,length=1):
    list_obj = list_obj or []
    if len(list_obj) >= length:
        list_obj = list_obj[:length]
    return list_obj

def get_image_metadata(file_path):
    """Extract image metadata (dimensions, file size)."""
    try:
        with Image.open(file_path) as img:
            width, height = img.size
            file_size = get_file_size(file_path)
        return {
            "dimensions": {"width": width, "height": height},
            "file_size": round(file_size, 3)
        }
    except Exception as e:
        return {"dimensions": {"width": 0, "height": 0}, "file_size": 0}

def update_json_data(json_data,update_data,keys=None):
    if keys == True:
        values_string = ''
        for key,value in update_data.items():
            values_string+= f"{key} == {value}"

        logger.info(f"new_datas:{values_string}")
        keys = valid_keys
    
    for key,value in update_data.items():
        if keys:
            if key in keys:
                json_data[key] = json_data.get(key) or value 
        else:
            json_data[key] = json_data.get(key) or value 
    return json_data


def update_sitemap(video_data,
                   sitemap_path):
    with open(sitemap_path, 'a') as f:
        f.write(f"""
<url>
    <loc>{video_data.get('canonical_url')}</loc>
    <video:video>
        <video:title>{video_data.get('seo_title')}</video:title>
        <video:description>{video_data.get('seo_description')}</video:description>
        <video:thumbnail_loc>{video_data.get('thumbnail',{}).get('file_path',{})}</video:thumbnail_loc>
        <video:content_loc>{video_data.get('video_path')}</video:content_loc>
    </video:video>
</url>
""")

def prune_inputs(func, *args, **kwargs):
    """
    Adapt the provided args/kwargs to fit the signature of func.
    Returns (args, kwargs) suitable for calling func.
    """
    sig = inspect.signature(func)
    params = sig.parameters

    # Handle positional arguments
    new_args = []
    args_iter = iter(args)
    for name, param in params.items():
        if param.kind in (inspect.Parameter.POSITIONAL_ONLY,
                          inspect.Parameter.POSITIONAL_OR_KEYWORD):
            try:
                new_args.append(next(args_iter))
            except StopIteration:
                break
        elif param.kind == inspect.Parameter.VAR_POSITIONAL:
            # collect all remaining args
            new_args.extend(args_iter)
            break
        else:
            break

    # Handle keyword arguments
    new_kwargs = {}
    for name, param in params.items():
        if name in kwargs:
            new_kwargs[name] = kwargs[name]
        elif param.default is inspect.Parameter.empty and param.kind == inspect.Parameter.KEYWORD_ONLY:
            # Required keyword not provided
            raise TypeError(f"Missing required keyword argument: {name}")

    # Only include keywords func accepts
    accepted_names = {
        name for name, p in params.items()
        if p.kind in (inspect.Parameter.POSITIONAL_OR_KEYWORD,
                      inspect.Parameter.KEYWORD_ONLY)
    }
    new_kwargs = {k: v for k, v in new_kwargs.items() if k in accepted_names}

    return tuple(new_args), new_kwargs




def generate_file_id(path: str, max_length: int = 50) -> str:
    base = os.path.splitext(os.path.basename(path))[0]
    base = unicodedata.normalize('NFKD', base).encode('ascii', 'ignore').decode('ascii')
    base = base.lower()
    base = re.sub(r'[^a-z0-9]+', '-', base).strip('-')
    base = re.sub(r'-{2,}', '-', base)
    if len(base) > max_length:
        h = hashlib.sha1(base.encode()).hexdigest()[:8]
        base = base[: max_length - len(h) - 1].rstrip('-') + '-' + h
    return base

def _get_video_id(*args,**kwargs):
    info_data = kwargs.get('info_data',kwargs) or kwargs or {}
    info_dir = info_data.get('info_dir') or info_data.get('info_directory')
    video_id = info_data.get('video_id')
    video_path = info_data.get('video_path')
    if info_dir:
        video_id = os.path.basename(info_dir)
    if video_path:
        video_id = generate_file_id(video_path)
    if video_id:
        return video_id

def get_videos_path(directory = None, info_data = None):
    info_data = info_data or {}
    if info_data and directory == None:
        directory = info_data['output_dir']
    directory = directory or TEXT_DIR
    return directory

def get_video_basenames(directory = None, info_data = None):
    directory = get_videos_path(directory = None, info_data = None)
    directory_items = os.listdir(directory)
    return directory_items


def get_videos_paths(directory = None, info_data = None):
    directory = get_videos_path(directory = directory, info_data = info_data)
    video_basenames = get_video_basenames(directory = directory, info_data = directory)
    directory_items = [os.path.join(directory,basename) for basename in video_basenames]
    return directory_items


def get_thumbnails_dir(info_dir=None,**kwargs):
    video_info_dir = info_dir or get_video_info_dir(**kwargs)
    thumbnails_directory=os.path.join(video_info_dir,'thumbnails')
    os.makedirs(thumbnails_directory,exist_ok=True)
    return thumbnails_directory


def get_video_info_dir(**kwargs):
    video_id = _get_video_id(**kwargs)
    info_dir = make_dirs(TEXT_DIR,video_id)
    os.makedirs(info_dir,exist_ok=True)
    get_thumbnails_dir(info_dir)
    return info_dir


def get_video_info_path(**kwargs):
    info_dir = get_video_info_dir(**kwargs)
    info_path = os.path.join(info_dir,'info.json')
    return info_path


def get_video_info_data(**kwargs):
    info_data=kwargs.get('info_data',kwargs) or kwargs  or {}
    info_file_path = None
    if info_data and isinstance(info_data,str) and os.path.isdir(info_data):
        info_dir = info_data
        info_file_path = os.path.join(info_dir,'info.json')
    elif info_data and isinstance(info_data,str) and os.path.isfile(info_data):
        info_file_path = info_data
    else:
        info_file_path = get_video_info_path(**info_data)
    if os.path.isfile(info_file_path):
        info_data = safe_load_from_json(info_file_path)
        return info_data


def get_audio_path(**kwargs):
    info_dir = get_video_info_dir(**kwargs)
    audio_path = os.path.join(info_dir,'audio.wav')
    return audio_path


def get_video_basename(**kwargs):
    video_path = kwargs.get('video_path')
    if not video_path:
        info_data = get_video_info_data(**kwargs)
        video_path = info_data.get('video_path')
    if video_path:
        basename= os.path.basename(video_path)
        return basename

def get_seo_title(title=None, keywords=None, filename=None, title_length=70, description=None):
    """Construct SEO title with keyword priority."""
    primary_keyword = filename or (keywords[0] if keywords else "")
    print(primary_keyword)
    seo_title = f"{primary_keyword} - {title}"
    title_length= title_length or 70
    return get_from_list(seo_title, length=title_length)


def get_seo_description(description=None, keywords=None, keyword_length=3, desc_length=300):
    """Construct SEO description with keyword hints."""
    seo_desc = f"{description or ''} Explore {keywords or ''}"
    return get_from_list(seo_desc, length=desc_length)


def get_title_tags_description(
    title=None,
    keywords=None,
    summary=None,
    filename=None,
    title_length=None,
    summary_length=150,
    keyword_length=3,
    desc_length=300,
    description=None,
):
    """Return SEO title, keyword string, description, and filtered tags."""
    summary_desc = get_from_list(description, length=summary_length)
    keywords_str = ""
    seo_title = get_seo_title(title=title, keywords=keywords, filename=filename, title_length=title_length)

    if isinstance(keywords, list):
        keywords = get_from_list(keywords, length=keyword_length)
        if keywords and len(keywords) > 0 and isinstance(keywords[0], list):
            keywords = keywords[0]
        if keywords:
            kedomainywords_str = ", ".join(keywords)

    seo_description = eatAll(
        get_seo_description(summary_desc, keywords_str, keyword_length=keyword_length, desc_length=desc_length),["'",'"',' ','','	']
    )
    seo_tags = [kw for kw in (keywords or []) if kw.lower() not in ["video", "audio", "file"]]
    return seo_title, keywords_str, seo_description, seo_tags


def get_seo_data(video_path=None,
                 filename=None,
                 title=None,
                 summary=None,
                 description=None,
                 keywords=None,
                 thumbnails_dir=None,
                 thumbnail_paths=None,
                 whisper_result=None,
                 audio_path=None,
                 domain=None,
                 **kwargs):
    """
    Enrich video/image info dict with SEO fields, captions, thumbnails, whisper, schema markup.
    """
    

    # Title/filename normalization

    info = {}
    domain = domain or "https://typicallyoutliers.com"
    if not filename and video_path:
        basename = os.path.basename(video_path)
        filename, ext = os.path.splitext(basename)
    title = title or filename

    # SEO text
    seo_title, keywords_str, seo_description, seo_tags = get_title_tags_description(
        title=title,
        keywords=keywords,
        summary=summary,
        filename=filename,
        description=description
        )
    info["seo_data"] = {"seo_title": seo_title, "seo_description": seo_description, "seo_tags": seo_tags,"keywords_str":keywords_str}
    info["seo_data"]["thumbnail"] = {"file_path": None, "alt_text": ""}
    # Thumbnail defaults
    try:
        
        if thumbnail_paths:
            thumb_file = thumbnail_paths[0]
            thumb_base = os.path.basename(thumb_file)
            alt_text = os.path.splitext(thumb_base)[0]
            info["seo_data"]["thumbnail"] = {"file_path": thumb_file, "alt_text": alt_text}
        elif thumbnails_dir and os.path.isdir(thumbnails_dir):
            thumbs = os.listdir(thumbnails_dir)
            if thumbs and len(thumbs) >0 and isinstance(thumbs,list):
                thumb_file = thumbs[0]
                thumb_base = os.path.join(thumbnails_dir,thumb_file)
                alt_text = os.path.splitext(thumb_file)[0]
                info["seo_data"]["thumbnail"] = {"file_path": thumb_file, "alt_text": alt_text}
        if whisper_result.get("segments"):
            thumb_score = pick_optimal_thumbnail(whisper_result, keywords,thumbnail_paths, thumbnails_dir, info=info)
            if thumb_score:
                frame, score, matched_text = thumb_score
                info["seo_data"]["thumbnail"].update({
                    "file_path": os.path.join(thumbs_dir, frame),
                    "alt_text": get_from_list(matched_text, length=100),
                })
    except:
        pass

    # Audio duration
    dur_s, dur_fmt = get_audio_duration(audio_path)
    info["seo_data"]["duration_seconds"]=dur_s
    info["seo_data"]["duration_formatted"]=dur_fmt

    # Schema + social metadata
    info["seo_data"]["schema_markup"] = {
        "@context": "https://schema.org",
        "@type": "VideoObject",
        "name": info["seo_data"]["seo_title"],
        "description": info["seo_data"]["seo_description"],
        "thumbnailUrl": info["seo_data"]["thumbnail"]["file_path"],
        "duration": f"PT{int(dur_s // 60)}M{int(dur_s % 60)}S",
        "uploadDate": get_time_now_iso(),
        "contentUrl": video_path,
        "keywords": info["seo_data"]["seo_tags"],
    }
    info["seo_data"]["social_metadata"] = {
        "og:title": info["seo_data"]["seo_title"],
        "og:description": info["seo_data"]["seo_description"],
        "og:image": info["seo_data"]["thumbnail"]["file_path"],
        "og:video": video_path,
        "twitter:card": "player",
        "twitter:title": info["seo_data"]["seo_title"],
        "twitter:description": info["seo_data"]["seo_description"],
        "twitter:image": info["seo_data"]["thumbnail"]["file_path"],
    }

    # Misc
    info["seo_data"]["categories"] = info["seo_data"].get("category",{})
    info["seo_data"]["category"] = next(
        (v for k, v in info["seo_data"]["categories"].items() if k in " ".join(info["seo_data"]["seo_tags"] or "").lower()), "General"
    )
    info["seo_data"]["uploader"] = info["seo_data"].get("uploader","typicallyoutliers")
    info["seo_data"]["uploader"] = {"name": info["seo_data"]["uploader"], "url": domain}
    info["seo_data"]["publication_date"] = get_time_now_iso()
    info["seo_data"]["video_metadata"] = get_video_metadata(video_path)
    info["seo_data"]["canonical_url"] = domain

    # Sitemap update
##    update_sitemap(info, f"{os.path.dirname(info['info_dir'])}/../sitemap.xml")
    return info




from .init_imports import *
@dataclass(frozen=True)
class StorageConfig:
    videos_root: str
    documents_root: str
    registry_path: str
    js_runtime: str = "deno"  # or from env

    @classmethod
    def from_env(cls) -> "StorageConfig":
        missing = []
        default_videos_dir = get_default_videos_dir()
        registry_path = get_env_value("VIDEO_REGISTRY_PATH") or os.path.join(default_videos_dir,'video_registry.json')
        videos_root = get_env_value("VIDEOS_ROOT") or get_default_videos_dir()
        documents_root = get_env_value("DOCUMENTS_ROOT") or get_default_documents_dir()
        if not videos_root:
            missing.append("VIDEOS_ROOT")
        if not documents_root:
            missing.append("DOCUMENTS_ROOT")
        if not registry_path:
            missing.append("VIDEO_REGISTRY_PATH")
        if missing:
            raise EnvironmentError(
                f"Required env vars not set: {missing}. "
                f"Set them explicitly — no defaults are assumed."
            )
        return cls(
            videos_root=videos_root,
            documents_root=documents_root,
            registry_path=registry_path,
            js_runtime=get_env_value("YT_DLP_JS_RUNTIME") or "deno",
        )
@dataclass(frozen=True)
class SiteConfig:
    site_name: str
    domain: str
    root_url: str
    media_root: str
    base_image_url: str

    @classmethod
    def from_env(cls) -> "SiteConfig":
        site_name = get_env_value("SITE_NAME")
        domain = get_env_value("SITE_DOMAIN")
        root_url = get_env_value("ROOT_URL")
        media_root = get_env_value("MEDIA_ROOT")
        base_image_url = get_env_value("BASE_IMAGE_URL")

        missing = [
            k for k, v in {
                "SITE_NAME": site_name,
                "SITE_DOMAIN": domain,
                "ROOT_URL": root_url,
                "MEDIA_ROOT": media_root,
                "BASE_IMAGE_URL": base_image_url,
            }.items() if not v
        ]
        if missing:
            raise EnvironmentError(f"Required env vars not set: {missing}")

        return cls(
            site_name=site_name,
            domain=domain,
            root_url=root_url,
            media_root=media_root,
            base_image_url=base_image_url,
        )

    # derived values — computed, not stored
    @property
    def video_media_root(self) -> str:
        return os.path.join(self.media_root, "videos")

    @property
    def video_public_url(self) -> str:
        return f"{self.root_url}/videos"

    @property
    def title_variants(self) -> list[str]:
        return title_variants_from_domain(self.domain)


@dataclass(frozen=True)
class AppConfig:
    site: SiteConfig
    storage: StorageConfig

    @classmethod
    def from_env(cls) -> "AppConfig":
        return cls(
            site=SiteConfig.from_env(),
            storage=StorageConfig.from_env(),
        )

def get_videos_root(explicit: str | None = None, envPath: str | None = None) -> str:
    """
    Resolve the videos root with this priority:
      1) explicit arg
      2) env file (VIDEOS_ROOT or legacy DATA_DIRECTORY)
      3) process env (VIDEOS_ROOT or legacy DATA_DIRECTORY)
      4) hard default (/mnt/24T/media/DATA/videos)
    """
    # 1) explicit
    if explicit:
        os.makedirs(explicit, exist_ok=True)
        return explicit

    # 2) env file
    v_from_envfile = (
        get_env_value(key="VIDEOS_ROOT", path=envPath)
        or get_env_value(key="DATA_DIRECTORY", path=envPath)  # legacy
    )
    if v_from_envfile:
        os.makedirs(v_from_envfile, exist_ok=True)
        return v_from_envfile

    # 3) process env
    v_from_env = os.getenv("VIDEOS_ROOT") or os.getenv("DATA_DIRECTORY")  # legacy
    if v_from_env:
        os.makedirs(v_from_env, exist_ok=True)
        return v_from_env

    # 4) hard default
    os.makedirs(DEFAULT_VIDEOS_ROOT, exist_ok=True)
    return DEFAULT_VIDEOS_ROOT


def get_documents_root(explicit: str | None = None, envPath: str | None = None) -> str:
    """
    Same idea as get_videos_root, but for documents/registry.
    Priority: explicit -> env file DOCUMENTS_ROOT -> env DOCUMENTS_ROOT -> hard default.
    """
    if explicit:
        os.makedirs(explicit, exist_ok=True)
        return explicit

    d_from_envfile = get_env_value(key="DOCUMENTS_ROOT", path=envPath)
    if d_from_envfile:
        os.makedirs(d_from_envfile, exist_ok=True)
        return d_from_envfile

    d_from_env = os.getenv("DOCUMENTS_ROOT")
    if d_from_env:
        os.makedirs(d_from_env, exist_ok=True)
        return d_from_env

    os.makedirs(DEFAULT_DOCUMENTS_ROOT, exist_ok=True)
    return DEFAULT_DOCUMENTS_ROOT


def get_video_directory(key: str | None = None, envPath: str | None = None, videos_root: str | None = None) -> str:
    """
    Assure that a valid *videos* directory exists and return its path.

    Priority:
      - videos_root arg (if passed)
      - env file (VIDEOS_ROOT or legacy DATA_DIRECTORY)
      - process env (VIDEOS_ROOT or legacy DATA_DIRECTORY)
      - DEFAULT_VIDEOS_ROOT (/mnt/24T/media/DATA/videos)
    """
    # keep key/envPath for backward compatibility with callers that used an env file
    # but prefer the explicit arg if provided
    root = get_videos_root(explicit=videos_root, envPath=envPath)
    os.makedirs(root, exist_ok=True)
    logger.info(f"using videos root: {root}")
    return root



from .init_imports import *
import uuid
def generate_video_id(*args,**kwargs):
    return f"{uuid.uuid4().hex}"
def get_info_path_from_video_path(video_path):
    video_parts = get_file_parts(video_path)
    dirname = video_parts.get('dirname')
    return os.path.join(dirname,'info.json')
    
def get_media_id(url=None, video_id=None, video_path=None):
    if url:
        video_id = get_video_id(url)
    if video_path and not video_id:
        info_path = get_info_path_from_video_path(video_path)
        if os.path.isfile(info_path):
            info_data = safe_load_from_json(info_path)
            video_id = info_data.get('id')
        else:
            video_id = generate_video_id(video_path)
    return video_id
def get_file_data(filepath, default=None):
    if not os.path.isfile(filepath):
        os.makedirs(os.path.dirname(filepath) or ".", exist_ok=True)
        safe_dump_to_file(default or {}, filepath)
    return safe_read_from_file(file_path=filepath)

def load_json_keys(filepath, main_data=None, default=None):
    base = default or {}
    main = dict(main_data or {})
    try:
        disk = get_file_data(filepath, default=base)
        for k, v in (main or {}).items():
            if isinstance(v, dict):
                v.update(disk.get(k, {}))
            else:
                main[k] = disk.get(k, v)
        # Also merge any keys present on disk but missing in main
        for k, v in disk.items():
            main.setdefault(k, v)
        return main
    except Exception:
        return main or base
def _first_existing(*paths):
    for p in paths:
        if p and os.path.isdir(p):
            return p
    return None



# at top:
import portalocker,fasteners,os
from dataclasses import dataclass
from abstract_webtools import (
    VideoDownloader,
    download_image,
    get_thumbnails,
    optimize_video_for_safari,
    download_video,dl_video,
    get_video_info,
    get_corrected_url,
    get_video_id,
    for_dl_video
    )
from abstract_utilities import *
from abstract_hugpy import *
import numpy as np
from pydub import AudioSegment
from abstract_react import getInfo,get_meta_info
get_pydub("AudioSegment")



from abstract_database import (
    execute_query,
    insert_any_combo,
    fetch_any_combo_,
    update_any_combo,
    connectionManager,
    create_connection
)
from abstract_utilities import get_env_value

# Registry: single source of truth for DB availability
DB_AVAILABLE = False

ENV_VARS = [
    "ABSTRACT_DATABASE_USER",
    "ABSTRACT_DATABASE_PORT",
    "ABSTRACT_DATABASE_DBNAME",
    "ABSTRACT_DATABASE_HOST",
    "ABSTRACT_DATABASE_PASSWORD",
]

def init_db(env_vars=None):
    """
    Attempts to establish a DB connection and create the videos table.
    Returns True if successful, False if anything fails.
    Sets the module-level DB_AVAILABLE flag accordingly.
    """
    global DB_AVAILABLE
    env_vars = env_vars or ENV_VARS

    env_values = {key.split('_')[-1].lower(): get_env_value(key) for key in env_vars}

    missing = [k for k, v in env_values.items() if not v]
    if missing:
        print(f"[db] Missing env vars: {missing}. DB disabled.")
        DB_AVAILABLE = False
        return False

    try:
        create_connection(**env_values)
        execute_query("""
            CREATE TABLE IF NOT EXISTS videos (
                id SERIAL PRIMARY KEY,
                video_id TEXT UNIQUE NOT NULL,
                info JSONB,
                metadata JSONB,
                whisper JSONB,
                captions JSONB,
                thumbnails JSONB,
                total_info JSONB,
                aggregated JSONB,
                seodata JSONB,
                audio_path TEXT,
                audio_format TEXT,
                created_at TIMESTAMP DEFAULT NOW(),
                updated_at TIMESTAMP DEFAULT NOW()
            )
        """, fetch=False)
        DB_AVAILABLE = True
        return True
    except Exception as e:
        print(f"[db] Failed to initialize: {e}. DB disabled.")
        DB_AVAILABLE = False
        return False


def upsert_video(video_id: str, **fields):
    if not DB_AVAILABLE:
        return None
    existing = fetch_any_combo_(
        table_name="videos",
        search_map={"video_id": video_id},
    )
    if existing:
        return update_any_combo(
            table_name="videos",
            update_map={**fields, "updated_at": "NOW()"},
            search_map={"video_id": video_id},
        )
    return insert_any_combo(
        table_name="videos",
        insert_map={"video_id": video_id, **fields},
    )


def get_video_db_record(video_id: str, hide_audio: bool = True) -> dict | None:
    if not DB_AVAILABLE:
        return None
    rows = fetch_any_combo_(
        table_name="videos",
        search_map={"video_id": video_id},
    )
    if not rows:
        return None
    record = rows[0]
    if hide_audio and "audio" in record:
        record["audio"] = f"<{len(record['audio'])} bytes>" if record["audio"] else None
    return record



##



from .imports import *
from .registry import *
from .utils import *
def get_info_registry(
    config: StorageConfig=None,
    videos_root=None,
    documents_root=None,
    flat_layout: bool = False,
    **kwargs
    ):
    return infoRegistry(
        config=config,
        videos_root=videos_root,
        documents_root=documents_root,
        flat_layout= flat_layout,
        **kwargs
        )
def get_video_info(
    url=None,
    video_id=None,
    video_path=None,
    force_refresh=False,
    config: StorageConfig=None,
    videos_root=None,
    documents_root=None,
    flat_layout: bool = False,
    **kwargs
    ):
    info_registry = infoRegistry(
        config=config,
        videos_root=videos_root,
        documents_root=documents_root,
        flat_layout= flat_layout,
        **kwargs
        )
    
    return info_registry.get_video_info(
        url=url,
        video_id=video_id,
        video_path=video_path,
        force_refresh=force_refresh
        )
def update_video_info(
    data,
    url=None,
    video_id=None,
    video_path=None,
    force_refresh=False,
    config: StorageConfig=None,
    videos_root=None,
    documents_root=None,
    flat_layout: bool = False,
    **kwargs
    ):
    info_registry = infoRegistry(
        config=config,
        videos_root=videos_root,
        documents_root=documents_root,
        flat_layout= flat_layout,
        **kwargs
        )
    info_registry.edit_info(data, video_id=video_id, url=url, video_path=video_path)


def extract_audio(
    url=None,
    video_id=None,
    video_path=None,
    force_refresh=False
    ):
    video_info = get_video_info(
        url=url,
        video_id=video_id,
        video_path=video_path,
        force_refresh=force_refresh)
    
    audio_path = video_info.get('audio_path')
    video_path = info.get('video_path')
    audio_path = os.path.join(video_dir,'audio.wav')
    extract_audio_from_video(video_path=video_path, audio_path=audio_path)
def get_analysis(
    text,
    scope=None,
    summary_preset=None,
    keyword_preset=None,
    input_policy=None
    ):
    scope='page'
    summary_preset= summary_preset or "article"
    keyword_preset= keyword_preset or "seo"
    input_policy=input_policy or "allow"
    return analyze_media_text(
        text,
        scope=scope,
        summary_preset=summary_preset,
        keyword_preset=keyword_preset,
        input_policy=input_policy
    )
def assure_video_download(
    url=None,
    video_id=None,
    video_path=None,
    force_refresh=False
    ):
    video_info = get_video_info(
        url=url,
        video_id=video_id,
        video_path=video_path,
        force_refresh=force_refresh)
    download=False
    video_path = video_info.get('video_path')
    info_path = video_info.get('info_path')
    video_dir = video_info.get('video_dir')
    video_basename = video_info.get('basename')
    
    if not os.path.isfile(video_path):
        info = for_dl_video(url,download_directory=video_dir,video_basename=video_basename)
        safe_dump_to_json(data=info,file_path=info_path)
        if os.path.isfile(info_path) or os.path.isfile(video_path):
            if os.path.isfile(video_path):
                video_info['checklist']["info"]=True
            if os.path.isfile(video_path):
                video_info['checklist']["download"]=True
            update_video_info(video_info,url=url,
                                video_id=video_id,
                                video_path=video_path,
                                force_refresh=force_refresh)
            
    if not os.path.isfile(info_path):
        info = for_dl_video(url,download_directory=video_dir,output_filename=video_basename,download=download)
        safe_dump_to_json(data=info,file_path=info_path)
        video_info['checklist']["info"]=True
        update_video_info(video_info,
                          url=url,
                          video_id=video_id,
                          video_path=video_path,
                          force_refresh=force_refresh
                          )
    return video_info
def assure_audio(
        url=None,
        video_id=None,
        video_path=None,
        force_refresh=False
        ):

    video_info = assure_video_download(
        url=url,
        video_id=video_id,
        video_path=video_path,
        force_refresh=force_refresh)
    
    audio_path = video_info.get('audio_path')

    video_path = video_info.get('video_path')
    if not os.path.isfile(audio_path):
        extract_audio_from_video(video_path=video_path, audio_path=audio_path)
        if os.path.isfile(audio_path):
            video_info['checklist']["audio"]=True
            update_video_info(video_info,
                          url=url,
                          video_id=video_id,
                          video_path=video_path,
                          force_refresh=force_refresh
                          )
    return video_info
def assure_video_thumbnails(
        url=None,
        video_id=None,
        video_path=None,
        force_refresh=False
        ):

    video_info = assure_video_download(
        url=url,
        video_id=video_id,
        video_path=video_path,
        force_refresh=force_refresh)
    thumbnails_dir = video_info.get('thumbnails_dir')
    info_path = video_info.get('info_path')
    video_dir = video_info.get('video_dir')
    if not os.path.isdir(thumbnails_dir) or len(os.listdir(thumbnails_dir)) == 0:
        info = safe_load_from_json(info_path)
        get_thumbnails(video_dir,info)
        if os.path.isdir(thumbnails_dir) and len(os.listdir(thumbnails_dir)) != 0:
            video_info['checklist']["thumbnails"]=True
            update_video_info(video_info,
                              url=url,
                              video_id=video_id,
                              video_path=video_path,
                              force_refresh=force_refresh
                              )
    return video_info
def assure_video_transcription(
        url=None,
        video_id=None,
        video_path=None,
        force_refresh=False
        ):

    video_info = assure_audio(
        url=url,
        video_id=video_id,
        video_path=video_path,
        force_refresh=force_refresh)
    audio_path = video_info.get('audio_path')
    whisper_path = video_info.get('whisper_path')
    if not os.path.isfile(whisper_path):
        whisper_data = whisper_transcribe(audio_path=audio_path)
        if whisper_data:
            safe_dump_to_json(data=whisper_data,file_path=whisper_path)
            if os.path.isfile(whisper_path):
                video_info['checklist']["whisper"]=True
                update_video_info(video_info,
                              url=url,
                              video_id=video_id,
                              video_path=video_path,
                              force_refresh=force_refresh
                              )
    return video_info
def assure_video_metadata(
        url=None,
        video_id=None,
        video_path=None,
        force_refresh=False
        ):

    video_info = assure_video_transcription(
        url=url,
        video_id=video_id,
        video_path=video_path,
        force_refresh=force_refresh)
    whisper_path = video_info.get('whisper_path')
    metadata_path = video_info.get('metadata_path')
    if os.path.isfile(whisper_path) and not os.path.isfile(metadata_path):
        whisper_data = safe_load_from_json(whisper_path)
        if whisper_data:
            whisper_text = whisper_data.get('text')
            if whisper_text:
                metadata = get_analysis(whisper_text)
                if metadata:
                    metadata = metadata.to_dict()
                    safe_dump_to_json(data=metadata,file_path=metadata_path)
                    if os.path.isfile(metadata_path):
                        video_info['checklist']["metadata"]=True
                        update_video_info(video_info,
                                      url=url,
                                      video_id=video_id,
                                      video_path=video_path,
                                      force_refresh=force_refresh
                                      )
    return video_info
def assure_video_seodata(
        url=None,
        video_id=None,
        video_path=None,
        force_refresh=False
        ):

    video_info = assure_video_metadata(
            url=url,
            video_id=video_id,
            video_path=video_path,
            force_refresh=force_refresh)
    seodata_path = video_info.get('seodata_path')
    if not os.path.isfile(seodata_path):
        config = SiteConfig.from_env()
        DOMAIN = config.domain
        info_data["metadata"]=metadata
        href = f"{DOMAIN}/video/{video_id}"
        title = info.get('title')
        summary = metadata.get('summary')
        hashtags = metadata.get('keywords').get('hashtags')
        meta_keywords = metadata.get('keywords').get('meta_keywords')
        optimal_thumbnail = pick_optimal_thumbnail(transcription,meta_keywords,thumbnail_paths=thumbnails,directory=video_dir)
        
        page_data = get_page_data(
            title=title,
            href=href,
            summary=summary,
            keywords=meta_keywords.replace(', ',',').split(','),
            keywords_str=meta_keywords,
            thumbnail=optimal_thumbnail or thumbnails[0])
        safe_dump_to_json(data=page_data,file_path=seodata_path)
        video_info['checklist']["seodata"]=True
        update_video_info(video_info,
                                      url=url,
                                      video_id=video_id,
                                      video_path=video_path,
                                      force_refresh=force_refresh
                                      )
    return video_info
def get_video_download(
        url=None,
        video_id=None,
        video_path=None,
        force_refresh=False
        ):
    video_info = assure_video_download(
            url=url,
            video_id=video_id,
            video_path=video_path,
            force_refresh=force_refresh)
    info_path = video_info.get('info_path')
    return safe_load_from_json(info_path)
def get_video_audio(
        url=None,
        video_id=None,
        video_path=None,
        force_refresh=False
        ):
    video_info = assure_audio(
            url=url,
            video_id=video_id,
            video_path=video_path,
            force_refresh=force_refresh)
    return video_info.get('audio_path')
def get_video_thumbnails(
        url=None,
        video_id=None,
        video_path=None,
        force_refresh=False
        ):
    video_info = assure_video_thumbnails(
            url=url,
            video_id=video_id,
            video_path=video_path,
            force_refresh=force_refresh)
    thumbnails_dir = video_info.get('thumbnails_dir')
    return [os.path.join(thumbnails_dir,item) for item in os.listdir(thumbnails_dir) if item]
def get_video_transcription(
        url=None,
        video_id=None,
        video_path=None,
        force_refresh=False
        ):
    video_info = assure_video_transcription(
            url=url,
            video_id=video_id,
            video_path=video_path,
            force_refresh=force_refresh)
    whisper_path = video_info.get('whisper_path')
    return safe_load_from_json(whisper_path)
def get_video_metadata(
        url=None,
        video_id=None,
        video_path=None,
        force_refresh=False
        ):
    video_info = assure_video_metadata(
            url=url,
            video_id=video_id,
            video_path=video_path,
            force_refresh=force_refresh)
    metadata_path = video_info.get('metadata_path')
    return safe_load_from_json(metadata_path)
def get_video_seodata(
        url=None,
        video_id=None,
        video_path=None,
        force_refresh=False
        ):
    video_info = assure_video_metadata(
            url=url,
            video_id=video_id,
            video_path=video_path,
            force_refresh=force_refresh)
    seodata_path = video_info.get('seodata_path')
    return safe_load_from_json(seodata_path)
def get_all(
        url=None,
        video_id=None,
        video_path=None,
        force_refresh=False
        ):
    info_data = get_video_info(
                url=url,
                video_id=video_id,
                video_path=video_path,
                force_refresh=force_refresh)
    info = get_video_download(
                url=url,
                video_id=video_id,
                video_path=video_path,
                force_refresh=force_refresh)
    info_data['video_info'] = info
    video_id = info_data.get('video_id')
    video_path = info_data.get('video_path')
    video_dir = info_data.get('video_dir')
    info_path = info_data.get('info_path')
    thumbnails = get_video_thumbnails(
                url=url,
                video_id=video_id,
                video_path=video_path,
                force_refresh=force_refresh)
    info_data["thumbnail_paths"]=thumbnails
    transcription = get_video_transcription(
                url=url,
                video_id=video_id,
                video_path=video_path,
                force_refresh=force_refresh)
    info_data["transcription"]=transcription
    metadata = get_video_metadata(
                url=url,
                video_id=video_id,
                video_path=video_path,
                force_refresh=force_refresh)
    info_data["metadata"]=metadata
    seodata = get_video_seodata(
                url=url,
                video_id=video_id,
                video_path=video_path,
                force_refresh=force_refresh)
    info_data["seodata"]=seodata
    return info_data



from .imports import *
import uuid
def generate_video_id(*args,**kwargs):
    return f"{uuid.uuid4().hex}"
def get_registry_data(config):
    registry_path = config.registry_path
    return safe_load_from_json(registry_path)
def search_config_for_id(config,url=None, video_id=None, video_path=None):
    url = get_corrected_url(url)
    
    registry_data = get_registry_data(config)
    
    if video_id and registry_data.get(video_id):
        return video_id
    for video_id,values in registry_data.items():
        if (url and url == values.get('url')) or (video_path and values.get('video_path')):
            return video_id
    return get_media_id(url=url, video_id=video_id, video_path=video_path)     
def get_video_record(config,url=None, video_id=None, video_path=None, force_refresh=False, hide_audio=False):
    video_id = search_config_for_id(
        config,
        url=url,
        video_id=video_id,
        video_path=video_path
        )
    registry_data= get_registry_data(config)
    values = registry_data.get(video_id)
    if values:
        return values
    video_dir = os.path.join(config.videos_root,video_id)
    filename= 'video'
    ext='.mp4'
    basename = f"{filename}{ext}"
    
    if video_path:
        video_dir = os.path.dirname(video_path)
        basename = os.path.basename(video_path)
        filename,ext= os.path.splitext(basename)
    else:
        video_path = os.path.join(video_dir,basename)
    info_path = os.path.join(video_dir,'info.json')
    audio_path = os.path.join(video_dir,'audio.wav')
    thumbnails_dir = os.path.join(video_dir,'thumbnails')
    audio_path = os.path.join(video_dir,'audio.wav')
    whisper_path = os.path.join(video_dir,'whisper.json')
    metadata_path = os.path.join(video_dir,'metadata.json')
    seodata_path = os.path.join(video_dir,'seodata.json')
    values = {
        "video_id":video_id,
        "url":url,
        "basename":basename,
        "ext":ext,
        "filename":filename,
        "info_path":info_path,
        "video_path":video_path,
        "video_dir":video_dir,
        "audio_path":audio_path,
        "thumbnails_dir":thumbnails_dir,
        "whisper_path":whisper_path,
        "metadata_path":metadata_path,
        "seodata_path":seodata_path,
        "checklist":{
            "download":False,
            "info":False,
            "thumbnails":False,
            "audio":False,
            "whisper":False,
            "metadata":False,
            "seodata":False
            }
        }
  
    
    return values
def _save_registry_unlocked(self):
    # Caller holds write_lock (thread-safety); now add process-safety.
    tmp = f"{self.registry_path}.tmp"
    safe_dump_to_file(self.registry, tmp)
    # atomic replace still helps, but lock ensures we don't interleave writers
    with open(self.registry_path, "wb") as f:
        portalocker.lock(f, portalocker.LOCK_EX)
        with open(tmp, "rb") as src:
            f.write(src.read())
        f.flush()
        os.fsync(f.fileno())
        portalocker.unlock(f)
    os.remove(tmp)
def _write_registry(config, video_id, data):
    """Atomic write to flat JSON registry — never leaves a partial file."""
    registry_path = config.registry_path
    registry_dir = os.path.dirname(registry_path)
    registry_data = safe_load_from_json(registry_path) or {}
    registry_data[video_id] = data
    
    fd, tmp_path = tempfile.mkstemp(dir=registry_dir, suffix='.tmp')
    try:
        with os.fdopen(fd, 'w') as f:
            json.dump(registry_data, f, indent=2)
        os.replace(tmp_path, registry_path)  # atomic on POSIX, best-effort on Windows
    except Exception as e:
        os.unlink(tmp_path)
        raise
class infoRegistry(metaclass=SingletonMeta):
    def __init__(self, config: StorageConfig = None, videos_root=None, documents_root=None, flat_layout: bool = False, **kwargs):
        if not hasattr(self, "initialized"):
            self.initialized = True
            self.config = config or StorageConfig.from_env()
            self.videos_root = self.config.videos_root
            self.documents_root = self.config.documents_root
            self.registry_path = self.config.registry_path
            if not os.path.isfile(self.registry_path):
                safe_dump_to_json(data={}, file_path=self.registry_path)
            self.db_available = init_db()

    def _upsert(self, video_id, data: dict):
        _write_registry(self.config, video_id, data)
        if self.db_available:
            try:
                upsert_video(video_id, info=data)
            except Exception as e:
                log.debug("DB upsert skipped for %s: %s", video_id, e)

    def get_video_info(self, url=None, video_id=None, video_path=None, force_refresh=False):

        values = get_video_record(self.config,url=url,video_id=video_id,video_path=video_path,force_refresh=force_refresh, hide_audio=False)
        self._upsert(video_id, values)
        return values
    def edit_info(self, data, video_id=None, url=None, video_path=None):
        if not video_id:
            video_id = get_video_id(url) if url else generate_video_id(video_path)
        self._upsert(video_id, data)
        return self.get_video_info(video_id=video_id, force_refresh=True)
